package fr.bhprod.sga.terrain.object;

import com.jme.image.Texture;
import com.jme.math.Matrix4f;
import com.jme.math.Vector3f;
import com.jme.renderer.Renderer;
import com.jme.scene.TriMesh;
import com.jmex.effects.ProjectedTextureUtil;

import fr.bhprod.sga.terrain.Terrain;
import fr.bhprod.sga.terrain.layer.DecalUtil;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public class Decal {
	/**
	 * Name of the projected decal mesh.
	 */
	private String name = null;
	/**
	 * Up Vector of the projector.
	 */
	private Vector3f up = null;
	/**
	 * Position of the projector.
	 */
	private Vector3f projectorPos = null;
	/**
	 * Texture of the decal.
	 */
	private Texture texture = null;
	/**
	 * Size factor of the decal.
	 */
	private float size = 100f;
	/**
	 * Near value of the projector.
	 */
	private float near = 0f;
	/**
	 * Far value of the projector.
	 */
	private float far = 0f;
	/**
	 * Field of view value of the projector.
	 */
	private float fov = 0f;
	/**
	 * Aspect value of the projector.
	 */
	private float aspect = 0f;
	/**
	 * Mesh of the projected decal.
	 */
	private TriMesh mesh = null;
	/**
	 * Delta between lookAt and projectorPos Vectors. 
	 */
	private static final float LOOK_DELTA = 0.0001f;
	/**
	 * The minimum size factor.
	 */
	private static final float MIN_SCALE = 50f;
	/**
	 * The maximum size factor.
	 */
	private static final float MAX_SCALE = 150f;

	/**
	 * Getter.
	 * @return value
	 */
	public final float getAspect() {
		return aspect;
	}

	/**
	 * Setter.
	 * @param pAspect param
	 */
	public final void setAspect(final float pAspect) {
		aspect = pAspect;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final float getFar() {
		return far;
	}

	/**
	 * Setter.
	 * @param pFar param
	 */
	public final void setFar(final float pFar) {
		far = pFar;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final float getFov() {
		return fov;
	}

	/**
	 * Setter.
	 * @param pFov param
	 */
	public final void setFov(final float pFov) {
		fov = pFov;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final String getName() {
		return name;
	}

	/**
	 * Setter.
	 * @param pName param
	 */
	public final void setName(final String pName) {
		name = pName;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final float getNear() {
		return near;
	}

	/**
	 * Setter.
	 * @param pNear param
	 */
	public final void setNear(final float pNear) {
		near = pNear;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final Vector3f getProjectorPos() {
		return projectorPos;
	}


	/**
	 * Setter.
	 * @param pSize param
	 */
	public final void setSize(final float pSize) {
		size = pSize;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final float getSize() {
		return size;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final TriMesh getMesh() {
		return mesh;
	}

	/**
	 * Setter.
	 * @param pMesh param
	 */
	public final void setMesh(final TriMesh pMesh) {
		mesh = pMesh;
	}

	/**
	 * Setter.
	 * @param pProjectorPos param
	 */
	public final void setProjectorPos(final Vector3f pProjectorPos) {
		projectorPos = pProjectorPos;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final Texture getTexture() {
		return texture;
	}

	/**
	 * Setter.
	 * @param pTexture param
	 */
	public final void setTexture(final Texture pTexture) {
		texture = pTexture;
	}

	/**
	 * Getter.
	 * @return value
	 */
	public final Vector3f getUp() {
		return up;
	}

	/**
	 * Setter.
	 * @param pUp param
	 */
	public final void setUp(final Vector3f pUp) {
		up = pUp;
	}

	/**
	 * Return the look Vector of the projector.<BR>
	 * @return The calculated look Vector
	 */
	public final Vector3f getLookAt() {
		return (new Vector3f(projectorPos.x + LOOK_DELTA, 0, projectorPos.z - LOOK_DELTA));
	}

	/**
	 * Verify if the decal parameters are presents.<BR>
	 * It throws an IllegalStateException if one of the parameter is absent
	 */
	private void verifyParams() {
		if (texture == null || projectorPos == null || up == null) {
			throw new IllegalStateException("All the decals parameters must be presents");
		}
	}
	
	/**
	 * Rotate the decal.
	 * @param angle The angle to turn the decal (RAD)
	 */
	public final void rotate(final float angle, Terrain terrain, Renderer renderer) {
		verifyParams();

		Matrix4f mat = new Matrix4f();
		mat.fromAngleAxis(angle, Vector3f.UNIT_Y);
		up.set(mat.mult(up));

		updateDecal(terrain, renderer);

		ProjectedTextureUtil.updateProjectedTexture(texture, fov, aspect, near,
				far, projectorPos, getLookAt(), up);
	}
	
	/**
	 * Scale the decal.
	 * @param value The factor to add to size
	 */
	public final void scale(final float value, Terrain terrain, Renderer renderer) {
		verifyParams();

		size += value;
		
		if (size <= MIN_SCALE) {
			size = MIN_SCALE;
		} else if (size >= MAX_SCALE) {
			size = MAX_SCALE;
		}
		
		projectorPos.y = size;
		
		updateDecal(terrain, renderer);

		ProjectedTextureUtil.updateProjectedTexture(texture, fov, aspect, near,
				far, projectorPos, getLookAt(), up);
	}

	/**
	 * Move the decal.
	 * @param x x move
	 * @param z z move
	 */
	public final void move(final float x, final float z, Terrain terrain, Renderer renderer) {
		verifyParams();

		// FIXME : Grer le dplacement indpendemment de la rotation de la camra
		projectorPos.set(projectorPos.add(new Vector3f(x, 0, z)));

		updateDecal(terrain, renderer);
	}

	/**
	 * Update the decal mesh in case of change.
	 */
	private void updateDecal(Terrain terrain, Renderer renderer) {
		// Recreate the mesh
		mesh.removeFromParent();
		mesh = null;
		DecalUtil.putDecal(this, terrain, renderer);
		
		ProjectedTextureUtil.updateProjectedTexture(texture, fov, aspect,
				near, far, projectorPos, getLookAt(), up);
	}
}

